require 'matrix'

module Rsb
  module Gol
    class Grid
      TOP_RIGHT_CORNER    = "┏"
      TOP_LEFT_CORNER     = "┓"
      BOTTOM_RIGHT_CORNER = "┛"
      BOTTOM_LEFT_CORNER  = "┛"
      HORIZONTAL_BAR      = "━"
      VERTICAL_BAR        = "┃"
      NEIGHBOR_NAMES      = [
        :north, :north_east, :north_west,
        :east, :west,
        :south, :south_east, :south_west
      ]

      attr_reader :rows, :columns, :top_border, :bottom_border

      def initialize(rows, columns, random_seed = false)
        @rows            = rows
        @columns         = columns
        @matrix          = Matrix.zero(rows, columns)
        hbar             = HORIZONTAL_BAR * columns
        @top_border      = TOP_RIGHT_CORNER + hbar + TOP_LEFT_CORNER
        @bottom_border   = BOTTOM_RIGHT_CORNER + hbar + BOTTOM_LEFT_CORNER

        apply_random_seed if random_seed == true
      end

      def spawn(x, y)
        fail "coordinates #{x},#{y} are out of bounds" if out_of_bounds?(x,y)
        @matrix.send(:[]=, x, y, 1)
      end

      def kill(x, y)
        fail "coordinates #{x},#{y} are out of bounds" if out_of_bounds?(x,y)
        @matrix.send(:[]=, x, y, 0)
      end

      def cell(x, y)
        @matrix[x, y]
      end

      def apply_random_seed
        @matrix.each_with_index do |state, x, y|
          spawn(x, y) if Random.rand(2) == 1
        end
      end

      def out_of_bounds?(x, y)
        x < 0 || y < 0 || x > rows - 1 || y > columns - 1
      end

      # Collects all live cells in a given neighborhood. A neighborhood is
      # defined as all other cells within on cell from the coordinates given.
      # At most a cell can have neighborhood of 8 cells, boundry cells will only
      # have 3 cell neighborhoods
      def neighborhood(x, y)
        neighbors = []
        NEIGHBOR_NAMES.each do |name|
          nx, ny = neighbor_coordinates(name, x, y)
          next if out_of_bounds?(nx, ny)
          state = cell(nx, ny)
          neighbors << state if state == 1
        end

        neighbors
      end

      def neighbor_coordinates(neighbor, x, y)
        case neighbor
        when :north       then [x - 1, y]
        when :north_east  then [x - 1, y - 1]
        when :north_west  then [x - 1, y + 1]
        when :east        then [x, y - 1]
        when :west        then [x, y + 1]
        when :south       then [x + 1, y]
        when :south_east  then [x + 1, y - 1]
        when :south_west  then [x + 1, y + 1]
        end
      end

      def cell_neighbors
         @matrix.each_with_index do |state, x, y|
            yield(state, x, y, neighborhood(x,y))
         end
      end

      def visualize(alive = 'o', dead = ' ')
        body = ''
        vbar = VERTICAL_BAR
        line = ''
        @matrix.each_with_index do |state, x, y|
          line << (state == 1 ? alive : dead)
          if y + 1 == columns
            body << "#{vbar}#{line}#{vbar}\n"
            line = ''
          end
        end
        "#{top_border}\n#{body}#{bottom_border}\n"
      end
    end
  end
end
